"""Servlet factory for unknown file types."""

import os

from mimetypes import guess_type

import HTTPExceptions

from HTTPServlet import HTTPServlet
from MiscUtils.Configurable import Configurable
from ServletFactory import ServletFactory

debug = False


try:
    from mimetypes import init as init_mimetypes
except ImportError:
    pass
else:  # workaround for Python issue #5853
    init_mimetypes()


class UnknownFileTypeServletFactory(ServletFactory):
    """The servlet factory for unknown file types.

    I.e. all files other than .py, .psp and the other types we support.
    """

    def uniqueness(self):
        return 'file'

    def extensions(self):
        return ['.*']

    def servletForTransaction(self, transaction):
        return UnknownFileTypeServlet(transaction.application())

    def flushCache(self):
        pass


# A cache of the files served up by UnknownFileTypeServlet cached by
# absolute, server side path. Each content is another dictionary with keys:
# content, mimeType, mimeEncoding.
# Previously, this content was stored directly in the attributes of the
# UnknownFileTypeServlets, but with that approach subclasses cannot
# dynamically serve content from different locations.
fileCache = {}


class UnknownFileTypeServlet(HTTPServlet, Configurable):
    """Servlet for unknown file types.

    Normally this class is just a "private" utility class for Webware's
    purposes. However, you may find it useful to subclass on occasion,
    such as when the server side file path is determined by something
    other than a direct correlation to the URL. Here is such an example:

    from UnknownFileTypeServlet import UnknownFileTypeServlet
    import os

    class Image(UnknownFileTypeServlet):

        imageDir = '/var/images'

        def filename(self, trans):
            filename = trans.request().field('i')
            filename = os.path.join(self.imageDir, filename)
            return filename
    """

    # region Candidates for subclass overrides

    def filename(self, trans):
        """Return the filename to be served.

        A subclass could override this in order to serve files from other
        disk locations based on some logic.
        """
        filename = getattr(self, '_serverSideFilename', None)
        if filename is None:
            filename = trans.request().serverSidePath()
            self._serverSideFilename = filename  # cache it
        return filename

    def shouldCacheContent(self):
        """Return whether the content should be cached or not.

        Returns a boolean that controls whether or not the content served
        through this servlet is cached. The default behavior is to return
        the CacheContent setting. Subclasses may override to always True
        or False, or incorporate some other logic.
        """
        return self.setting('CacheContent')

    # endregion Candidates for subclass overrides

    # region Init et al

    def __init__(self, application):
        HTTPServlet.__init__(self)
        Configurable.__init__(self)
        self._application = application

    def defaultConfig(self):
        """Get the default config.

        Taken from Application's 'UnknownFileTypes' default setting.
        """
        return self._application.defaultConfig()['UnknownFileTypes']

    def userConfig(self):
        """Get the user config.

        Taken from Application's 'UnknownFileTypes' user setting.
        """
        return self._application.userConfig().get('UnknownFileTypes', {})

    def configFilename(self):
        return self._application.configFilename()

    def canBeReused(self):
        return self.setting('ReuseServlets')

    @staticmethod
    def validTechniques():
        return ['serveContent', 'redirectSansScript']

    def respondToGet(self, trans):
        """Respond to GET request.

        Responds to the transaction by invoking self.foo() for foo is
        specified by the 'Technique' setting.
        """
        technique = self.setting('Technique')
        if technique == 'redirectSansAdapter':
            oldTechnique, technique = technique, 'redirectSansScript'
            print(f'The {oldTechnique!r} technique is deprecated,'
                  f' please use {technique!r} instead')
        if technique not in self.validTechniques():
            raise AttributeError(f'Invalid technique: {technique!r}')
        method = getattr(self, technique)
        method(trans)

    respondToHead = respondToGet

    def respondToPost(self, trans):
        """Respond to POST request.

        Invoke self.respondToGet().

        Since posts are usually accompanied by data, this might not be the
        best policy. However, a POST would most likely be for a CGI, which
        currently no one is mixing in with their Webware-based web sites.
        """
        self.respondToGet(trans)

    @staticmethod
    def redirectSansScript(trans):
        """Redirect to web server.

        Sends a redirect to a URL that doesn't contain the script name.
        Under the right configuration, this will cause the web server to
        then be responsible for the URL rather than the WSGI server.
        Keep in mind that links off the target page will *not* include
        the script name in the URL.
        """
        request = trans.request()
        newURL = os.path.split(request.scriptName())[0] + request.pathInfo()
        newURL = newURL.replace('//', '/')
        trans.response().sendRedirect(newURL)

    def lastModified(self, trans):
        try:
            return os.path.getmtime(self.filename(trans))
        except OSError:
            return None

    def serveContent(self, trans):
        response = trans.response()

        maxCacheContentSize = self.setting('MaxCacheContentSize')
        readBufferSize = self.setting('ReadBufferSize')

        # start sending automatically
        response.streamOut().setAutoCommit()

        filename = self.filename(trans)
        try:
            f = open(filename, 'rb')
        except IOError as e:
            raise HTTPExceptions.HTTPNotFound from e

        stat = os.fstat(f.fileno())
        fileSize, mtime = stat[6], stat[8]

        if debug:
            print('>> UnknownFileType.serveContent()')
            print('>> filename =', filename)
            print('>> size=', fileSize)
        fileDict = fileCache.get(filename)
        if fileDict is not None and mtime != fileDict['mtime']:
            # Cache is out of date; clear it.
            if debug:
                print('>> changed, clearing cache')
            del fileCache[filename]
            fileDict = None
        if fileDict is None:
            if debug:
                print('>> not found in cache')
            mimeType, mimeEncoding = guess_type(filename, False)
            if mimeType is None:
                mimeType, mimeEncoding = 'application/octet-stream', None
        else:
            mimeType = fileDict['mimeType']
            mimeEncoding = fileDict['mimeEncoding']
        response.setHeader('Content-Type', mimeType)
        response.setHeader('Content-Length', str(fileSize))
        if mimeEncoding:
            response.setHeader('Content-Encoding', mimeEncoding)
        if trans.request().method() == 'HEAD':
            f.close()
            return
        if (fileDict is None and self.setting('ReuseServlets')
                and self.shouldCacheContent()
                and fileSize < maxCacheContentSize):
            if debug:
                print('>> caching')
            fileDict = dict(
                content=f.read(),
                mimeType=mimeType, mimeEncoding=mimeEncoding,
                mtime=mtime, size=fileSize, filename=filename)
            fileCache[filename] = fileDict
        if fileDict is not None:
            if debug:
                print('>> sending content from cache')
            response.write(fileDict['content'])
        else:  # too big or not supposed to cache
            if debug:
                print('>> sending directly')
            numBytesSent = 0
            while numBytesSent < fileSize:
                data = f.read(min(fileSize-numBytesSent, readBufferSize))
                if data == '':
                    break  # unlikely, but safety first
                response.write(data)
                numBytesSent += len(data)
        f.close()

    # endregion Init et al



"""RPC servlets."""

import traceback
import sys

from HTTPServlet import HTTPServlet


class RPCServlet(HTTPServlet):
    """RPCServlet is a base class for RPC servlets."""

    def call(self, methodName, *args, **keywords):
        """Call custom method.

        Subclasses may override this class for custom handling of methods.
        """
        if methodName not in self.exposedMethods():
            raise NotImplementedError(methodName)
        return getattr(self, methodName)(*args, **keywords)

    @staticmethod
    def exposedMethods():
        """Get exposed methods.

        Subclasses should return a list of methods that will be exposed
        through XML-RPC.
        """
        return ['exposedMethods']

    def resultForException(self, e, trans):
        """Get text for exception.

        Given an unhandled exception, returns the string that should be
        sent back in the RPC response as controlled by the
        RPCExceptionReturn setting.
        """
        # report exception back to server
        setting = trans.application().setting('RPCExceptionReturn')
        if setting == 'occurred':
            result = 'unhandled exception'
        elif setting == 'exception':
            result = str(e)
        elif setting == 'traceback':
            result = ''.join(traceback.format_exception(*sys.exc_info()))
        else:
            raise ValueError(f'Invalid setting: {setting!r}')
        return result

    @staticmethod
    def sendOK(contentType, contents, trans, contentEncoding=None):
        """Send a 200 OK response with the given contents."""
        response = trans.response()
        response.setStatus(200, 'OK')
        response.setHeader('Content-Type', contentType)
        response.setHeader('Content-Length', str(len(contents)))
        if contentEncoding:
            response.setHeader('Content-Encoding', contentEncoding)
        response.write(contents)

    @staticmethod
    def handleException(transaction):
        """Handle exception.

        If ReportRPCExceptionsInWebware is set to True, then flush the response
        (because we don't want the standard HTML traceback to be appended to
        the response) and then handle the exception in the standard Webware
        way. This means logging it to the console, storing it in the error log,
        sending error email, etc. depending on the settings.
        """
        setting = transaction.application().setting(
            'ReportRPCExceptionsInWebware')
        if setting:
            transaction.response().flush()
            transaction.application().handleExceptionInTransaction(
                sys.exc_info(), transaction)

    def transaction(self):
        """Get the corresponding transaction.

        Most uses of RPC will not need this.
        """
        return self._transaction

    def awake(self, transaction):
        """Begin transaction."""
        HTTPServlet.awake(self, transaction)
        self._transaction = transaction

    def sleep(self, transaction):
        """End transaction."""
        self._transaction = None
        HTTPServlet.sleep(self, transaction)



"""Session store using the shelve module."""

import os
import shelve
import threading

from MiscUtils import NoDefault

from SessionStore import maxPickleProtocol, SessionStore


class SessionShelveStore(SessionStore):
    """A session store implemented with a shelve object.

    To use this store, set SessionStore in Application.config to 'Shelve'.
    """

    _filename = 'Session.Store'

    # region Init

    def __init__(self, app, restoreFiles=None, filename=None):
        """Initialize the session shelf.

        If restoreFiles is true, existing shelve file(s) will be reused.
        """
        SessionStore.__init__(self, app)
        filename = os.path.join(app._sessionDir, filename or self._filename)
        if restoreFiles is None:
            restoreFiles = self._retain
        flag = 'c' if restoreFiles else 'n'
        self._store = shelve.open(
            filename, flag=flag, protocol=maxPickleProtocol)
        self._lock = threading.RLock()

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions."""
        return len(self._store)

    def __getitem__(self, key):
        """Get a session item, reading it from the store."""
        # multiple simultaneous read accesses are safe
        return self._store[key]

    def __setitem__(self, key, value):
        """Set a session item, writing it to the store."""
        # concurrent write access is not supported
        dirty = value.isDirty()
        if self._alwaysSave or dirty:
            with self._lock:
                if dirty:
                    value.setDirty(False)
                try:
                    self._store[key] = value
                except Exception:
                    if dirty:
                        value.setDirty()
                    raise  # raise original exception

    def __delitem__(self, key):
        """Delete a session item from the store."""
        with self._lock:
            session = self[key]
            if not session.isExpired():
                session.expiring()
            del self._store[key]

    def __contains__(self, key):
        """Check whether the session store has a given key."""
        return key in self._store

    def __iter__(self):
        """Return an iterator over the stored session keys."""
        return iter(self._store)

    def keys(self):
        """Return a list with the keys of all the stored sessions."""
        return list(self._store.keys())

    def clear(self):
        """Clear the session store, removing all of its items."""
        self._store.clear()

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        with self._lock:
            return self._store.setdefault(key, default)

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        with self._lock:
            if default is NoDefault:
                return self._store.pop(key)
            return self._store.pop(key, default)

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save potentially changed session in the store."""
        key = session.identifier()
        if key not in self or self[key] is not session:
            self[key] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store.

        Should be used (only) when the application server is shut down.
        """
        self._store.close()

    def cleanStaleSessions(self, task=None):
        """Clean stale sessions."""
        SessionStore.cleanStaleSessions(self, task)
        self.intervalSweep()

    def intervalSweep(self):
        """The session sweeper interval function."""
        self._store.sync()

    # endregion Application support



import os
import time

from URLParser import ServletFactoryManager
from WebUtils.Funcs import htmlEncode
from .AdminSecurity import AdminSecurity


class ServletCache(AdminSecurity):
    """Display servlet cache.

    This servlet displays, in a readable form, the internal data
    structure of the cache of all servlet factories.

    This can be useful for debugging Webware problems and the
    information is interesting in general.
    """

    def title(self):
        return 'Servlet Cache'

    def writeContent(self):
        wr = self.writeln
        factories = [
            factory for factory in ServletFactoryManager._factories
            if factory._classCache]
        if not factories:
            wr('<h4>No caching servlet factories found.</h4>')
            wr('<p>Caching can be activated by setting'
               ' <code>CacheServletClasses = True</code>.</p>')
        if len(factories) > 1:
            factories.sort()
            wr('<h3>Servlet Factories:</h3>')
            wr('<table>')
            for factory in factories:
                name = factory.name()
                wr('<tr><td><a href="#{0}">{0}</a></td></tr>'.format(name))
            wr('</table>')
        hasField = self.request().hasField
        wr('<form action="ServletCache" method="post">')
        for factory in factories:
            name = factory.name()
            wr('<a id="{0}"></a><h4>{0}</h4>'.format(name))
            if hasField('flush_' + name):
                factory.flushCache()
                wr('<p style="color:green">'
                   'The servlet cache has been flushed. &nbsp; '
                   '<input type="submit" name="reload" value="Reload"></p>')
                continue
            wr(htCache(factory))
        wr('</form>')


def htCache(factory):
    """Output the cache of a servlet factory."""
    html = []
    wr = html.append
    cache = factory._classCache
    keys = sorted(cache)
    wr(f'<p>Uniqueness: {factory.uniqueness()}</p>')
    extensions = ', '.join(map(repr, factory.extensions()))
    wr(f'<p>Extensions: {extensions}</p>')
    name = factory.name()
    wr(f'<p>Unique paths in the servlet cache: <strong>{len(keys)}</strong>'
       f' &nbsp; <input type="submit" name="flush_{name}" value="Flush"></p>')
    wr('<p>Click any link to jump to the details for that path.</p>')
    wr('<h5>Filenames:</h5>')
    wr('<table class="NiceTable">')
    wr('<tr><th>File</th><th>Directory</th></tr>')
    paths = []
    for key in keys:
        head, tail = os.path.split(key)
        path = dict(dir=head, base=tail, full=key, id=id(key))
        paths.append(path)
    paths.sort(key=lambda p: (p['base'].lower(), p['dir'].lower()))
    for path in paths:
        wr('<tr><td><a href="#id{id}">{base}</a></td>'
           '<td>{dir}</td></tr>'.format(**path))
    wr('</table>')
    wr('<h5>Full paths:</h5>')
    wr('<table class="NiceTable">')
    wr('<tr><th>Servlet path</th></tr>')
    for key in keys:
        wr(f'<tr><td><a href="#{id(key)}">{key}</a></td></tr>')
    wr('</table>')
    wr('<h5>Details:</h5>')
    wr('<table class="NiceTable">')
    for path in paths:
        wr('<tr class="NoTable"><td colspan="2"><a id="id{id}"></a>'
           '<strong>{base}</strong> - {dir}</td></tr>'.format(**path))
        record = cache[path['full']].copy()
        record['path'] = path['full']
        record['instances'] = (
            'one servlet instance (threadsafe)'
            if path['full'] in factory._threadsafeServletCache else
            f'free reusable servlets: {len(factory._servletPool)}')
        wr(htRecord(record))
    wr('</table>')
    return '\n'.join(html)


def htRecord(record):
    html = []
    wr = html.append
    for key in sorted(record):
        htKey = htmlEncode(key)
        # determine the HTML for the value
        value = record[key]
        htValue = None
        # check for special cases where we want a custom display
        if hasattr(value, '__name__'):
            htValue = value.__name__
        if key == 'mtime':
            htValue = time.asctime(time.localtime(value))
            htValue = f'{htValue} ({value})'
        # the general case:
        if not htValue:
            htValue = htmlEncode(str(value))
        wr(f'<tr><th>{htKey}</th><td>{htValue}</td></tr>')
    return '\n'.join(html)



"""Default base class for PSP pages.

This class is intended to be used in the future as the default base class
for PSP pages in the event that some special processing is needed.
Right now, no special processing is needed, so the default base class
for PSP pages is the standard Webware Page.
"""

from Page import Page


class PSPPage(Page):

    def __init__(self):
        self._parent = Page
        self._parent.__init__(self)

    def awake(self, transaction):
        self._parent.awake(self, transaction)
        self.out = transaction.response()



import sys

from importlib import reload

from .AdminSecurity import AdminSecurity


class AppControl(AdminSecurity):

    def writeContent(self):
        req = self.request()
        wr = self.writeln
        action = self.request().field("action", None)

        if action is None:
            wr('''<form action="AppControl" method="post">
<table>
<tr><td><input type="submit" name="action" value="Clear cache"></td>
<td>Clear the class and instance caches of each servlet factory.</td>
</tr><tr>
<td><input type="submit" name="action" value="Reload"></td>
<td>Reload the selected Python modules. Be careful!</td></tr>''')
            wr('<tr><td></td><td>')
            for n in sorted(sys.modules):
                m = sys.modules[n]
                if (not n.endswith('__init__') and not hasattr(m, '__path__')
                        and not hasattr(m, '__orig_file__')):
                    # show only the easily reloadable modules
                    wr(f'<input type="checkbox" name="reloads" value="{n}">'
                       f' {n}<br>')
            wr('</td></tr>\n</table>\n</form>')

        elif action == "Clear cache":
            from URLParser import ServletFactoryManager
            factories = [f for f in ServletFactoryManager._factories
                         if f._classCache]
            wr('<p>')
            for factory in factories:
                wr(f'Flushing cache of {factory.name()}...<br>')
                factory.flushCache()
            wr('</p>')
            wr('<p style="color:green">The caches of all factories'
               ' have been flushed.</p>')
            wr('<p>Click here to view the Servlet cache:'
               ' <a href="ServletCache">Servlet Cache</a></p>')

        elif action == "Reload":
            wr('<p>Reloading selected modules. Any existing classes'
               ' will continue to use the old module definitions,'
               ' as will any functions/variables imported using "from".'
               ' Use "Clear Cache" to clean out any servlets'
               ' in this condition.<p>')
            reloadNames = req.field("reloads", None)
            if not isinstance(reloadNames, list):
                reloadNames = [reloadNames]
            wr('<p>')
            for n in reloadNames:
                m = sys.modules.get(n)
                if m:
                    wr(f"Reloading {self.htmlEncode(str(m))}...<br>")
                    try:
                        reload(m)
                    except Exception as e:
                        wr('<span style="color:red">Could not reload, '
                           f'error was "{e}".</span><br>')
            wr('</p>')
            wr('<p style="color:green">The selected modules'
               ' have been reloaded.</p>')

        else:
            wr(f'<p>Cannot perform "{action}".</p>')



"""Test Webware Application via its WSGI interface"""

import sys

from io import StringIO
from os import chdir, getcwd

from webtest import TestApp, AppError  # pylint: disable=import-error

from Application import Application

__all__ = ['AppTest', 'AppError']


class AppTest:

    settings = dict(
        PrintConfigAtStartUp=False
    )
    catchOutput = True  # set to False if you want to run pdb inside a test

    @classmethod
    def setUpClass(cls):
        stdout, stderr = sys.stdout, sys.stderr
        if cls.catchOutput:
            sys.stdout = sys.stderr = StringIO()
        cls.currentDir = getcwd()
        import webware
        webwareDir = webware.__path__[0]
        chdir(webwareDir)
        try:
            app = Application(settings=cls.settings, development=True)
            cls.app = app
            cls.testApp = TestApp(app)
        except Exception as e:
            error = str(e) or 'Could not create application'
        else:
            error = None
        finally:
            if cls.catchOutput:
                output = sys.stdout.getvalue().rstrip()
                sys.stdout, sys.stderr = stdout, stderr
            else:
                output = ''
        if error:
            raise RuntimeError(
                'Error setting up application:\n' + error +
                '\nOutput was:\n' + output)
        if cls.catchOutput and not (
                output.startswith('Webware for Python')
                and 'Running in development mode' in output
                and 'Loading context' in output):
            raise AssertionError(
                'Application was not properly started.'
                ' Output was:\n' + output)

    @classmethod
    def tearDownClass(cls):
        stdout, stderr = sys.stdout, sys.stderr
        if cls.catchOutput:
            sys.stdout = sys.stderr = StringIO()
        cls.app.shutDown()
        if cls.catchOutput:
            output = sys.stdout.getvalue().rstrip()
            sys.stdout, sys.stderr = stdout, stderr
        else:
            output = ''
        chdir(cls.currentDir)
        if cls.catchOutput and output != (
                'Application is shutting down...\n'
                'Application has been successfully shut down.'):
            raise AssertionError(
                'Application was not properly shut down. Output was:\n'
                + output)

    def setUp(self):
        self.stdout, self.stderr = sys.stdout, sys.stderr
        if self.catchOutput:
            sys.stdout = sys.stderr = StringIO()

    def tearDown(self):
        if self.catchOutput:
            self.output = sys.stdout.getvalue().rstrip()
            sys.stdout, sys.stderr = self.stdout, self.stderr
        else:
            self.output = ''

    def run(self, result=None):
        result = super().run(result)  # pylint: disable=no-member
        if not result.wasSuccessful() and self.output:
            print("Application output was:")
            print(self.output)



try:
    from yattag import Doc
except ImportError:
    Doc = None

from .ExamplePage import ExamplePage


class YattagDemo(ExamplePage):
    """Demo for using the Yattag library."""

    defaults = {
        'subject': 'Just testing',
        'message': 'You just won the lottery!'
    }

    spamWords = 'earn free income lottery money offer price'.split()

    def isSpam(self, text):
        text = text.lower()
        for word in self.spamWords:
            if word in text:
                return True
        return False

    def writeContent(self):
        self.writeln('<h1>Using Webware with Yattag</h1>')
        self.writeln(
            '<p>Yattag is a Python library that can be used in Webware'
            ' applications to generate HTML programmatically.</p>')
        if Doc:
            self.writeln(self.demoForm())
        else:
            self.writeln(
                '<p>Please <a href="https://www.yattag.org/download-install">'
                ' install yattag</a> in order to view this demo.</p>')

    def demoForm(self):
        defaults = self.defaults.copy()
        errors = {}
        request = self.request()
        if request.hasField('message'):
            subject = self.request().field('subject')
            defaults['subject'] = subject
            if not subject:
                errors['subject'] = 'Please add the subject of your message.'
            elif self.isSpam(subject):
                errors['subject'] = 'Your subject looks like spam!'
            message = self.request().field('message')
            defaults['message'] = message
            if not message:
                errors['message'] = 'You did not enter a message!'
            elif self.isSpam(message):
                errors['message'] = 'Your message looks like spam!'
        else:
            subject = message = None

        doc, tag, text = Doc(defaults=defaults, errors=errors).tagtext()

        if message and not errors:

            with tag('p', klass='success'):
                text('Congratulations! You have sent the following message:')
            with tag('div', klass='output'):
                with tag('p'):
                    with tag('strong'):
                        text(subject)
                with tag('p'):
                    text(message)
            with tag('a', href='YattagDemo'):
                text('Try sending another message.')

        else:

            with tag('h2'):
                text('Demo contact form')

            with tag('form', action="YattagDemo"):
                doc.input(name='subject', type='text', size=80)
                with doc.textarea(
                        name='message', cols=80, rows=8):
                    pass
                doc.stag('input', type='submit', value='Send my message')

        return doc.getvalue()

    def writeStyleSheet(self):
        super().writeStyleSheet()
        if Doc:
            style = """
#Content form input, #Content form textarea {
    display: block; margin: 1em 0; padding: 2pt;
}
#Content form input.error, #Content form textarea.error {
    border: 1px solid red !important;
    box-shadow: 0 0 3px red important;
}
#Content form span.error {
    color: red;
}
#Content p.success {
    color: green;
}
#Content .output {
margin: 2em;
padding: 0.5em 1em;
border: 1px solid #cce;
}
            """
            doc, tag, text = Doc().tagtext()
            with tag('style'):
                text(style)

            self.writeln(doc.getvalue())



"""Mock Webware Application class."""


class Application:
    """Mock application."""

    _sessionDir = 'SessionStoreTestDir'
    _alwaysSaveSessions = True
    _retainSessions = True
    _sessionTimeout = 60 * 60
    _sessionPrefix = ''
    _sessionName = '_SID_'

    def setting(self, key, default=None):
        return dict(
            SessionTimeout=self._sessionTimeout // 60,
            SessionPrefix=self._sessionPrefix,
            SessionName=self._sessionName,
            DynamicSessionTimeout=1,
            MaxDynamicMemorySessions=3,
            MemcachedOnIteration=None,
            Debug=dict(Sessions=False),
        ).get(key, default)

    def handleException(self):
        raise Exception('Application Error')

    def sessionTimeout(self, _trans=None):
        return self._sessionTimeout

    def sessionPrefix(self, _trans=None):
        return self._sessionPrefix

    def sessionName(self, _trans=None):
        return self._sessionName

    def hasSession(self, _sessionId):
        return False



"""Test Webware Testing context"""

import unittest

from .AppTest import AppTest


class TestContextMap(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        ExtraPathInfo=False,
        Contexts={'Testing': 'Testing', 'default': 'Testing'}
    )

    def testStartPage(self):
        r = self.testApp.get('/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/')
        r.mustcontain('<title>Testing</title>')

    def testContextPage(self):
        r = self.testApp.get('/Testing/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Testing/')
        r.mustcontain('<title>Testing</title>')

    def testBadContextPage(self):
        r = self.testApp.get('/default/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)
        r = self.testApp.get('/TestAlias/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)


class TestAliasedContext(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        ExtraPathInfo=False,
        Contexts={'TestAlias': 'Testing', 'default': 'TestAlias'}
    )

    def testStartPage(self):
        r = self.testApp.get('/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/')
        r.mustcontain('<title>Testing</title>')

    def testContextPage(self):
        r = self.testApp.get('/TestAlias/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/TestAlias/')
        r.mustcontain('<title>Testing</title>')

    def testBadContextPage(self):
        r = self.testApp.get('/default/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)
        r = self.testApp.get('/Testing/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)


class TestOnlyDefaultContext(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        ExtraPathInfo=False,
        Contexts={'default': 'Testing'}
    )

    def testStartPage(self):
        r = self.testApp.get('/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/')
        r.mustcontain('<title>Testing</title>')

    def testContextPage(self):
        r = self.testApp.get('/default/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/default/')
        r.mustcontain('<title>Testing</title>')

    def testBadContextPage(self):
        r = self.testApp.get('/Testing/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)
        r = self.testApp.get('/TestAlias/', expect_errors=True)
        self.assertEqual(r.status, '404 Not Found', r.status)



"""Abstract servlets"""

import os

from MiscUtils import AbstractError
from MiscUtils.Funcs import asclocaltime


class Servlet:
    """A general servlet.

    A servlet is a key portion of a server-based application that implements
    the semantics of a particular request by providing a response.
    This abstract class defines servlets at a very high level.
    Most often, developers will subclass HTTPServlet or even Page.

    Servlets can be created once, then used and destroyed, or they may be
    reused several times over (it's up to the server). Therefore, servlet
    developers should take the proper actions in awake() and sleep()
    so that reuse can occur.

    Objects that participate in a transaction include:
      * Application
      * Request
      * Transaction
      * Session
      * Servlet
      * Response

    The awake(), respond() and sleep() methods form a message sandwich.
    Each is passed an instance of Transaction which gives further access
    to all the objects involved.
    """

    # region Init

    def __init__(self):
        """Subclasses must invoke super."""
        self._serverSidePath = None
        self._factory = None
        self._busy = False

    def __repr__(self):
        name = self.name()
        try:
            path = self.serverSidePath()
        except AttributeError:
            path = None
        if path:
            name += f' at {path}'
        return f'<Servlet {name}>'

    # endregion Init

    # region Access

    def name(self):
        """Return the name which is simple the name of the class.

        Subclasses should *not* override this method.
        It is used for logging and debugging.
        """
        return self.__class__.__name__

    # endregion Access

    # region Request-response cycles

    @staticmethod
    def runTransaction(transaction):
        try:
            transaction.awake()
            transaction.respond()
        except Exception as first:
            try:
                transaction.sleep()
            except Exception as second:
                # The first exception is more important than the *second* one
                # that comes from sleep(). In fact, without this little trick
                # the first exception gets hidden by the second which is often
                # just a result of the first. Then you're stuck scratching your
                # head wondering what the first might have been.
                raise second from first
            else:
                # no problems with sleep() so raise the one and only exception
                raise
        else:
            transaction.sleep()

    def runMethodForTransaction(self, transaction, method, *args, **kw):
        self.awake(transaction)
        result = getattr(self, method)(*args, **kw)
        self.sleep(transaction)
        return result

    def awake(self, transaction):
        """Send the awake message.

        This message is sent to all objects that participate in the
        request-response cycle in a top-down fashion, prior to respond().
        Subclasses must invoke super.
        """
        self._transaction = transaction

    def respond(self, transaction):
        """Respond to a request."""
        raise AbstractError(self.__class__)

    def sleep(self, transaction):
        """Send the sleep message."""
        # base method does nothing

    # endregion Request-response cycles

    # region Log

    def log(self, message):
        """Log a message.

        This can be invoked to print messages concerning the servlet.
        This is often used by self to relay important information back
        to developers.
        """
        print(f'[{asclocaltime()}] [msg] {message}')

    # endregion Log

    # region Abilities

    @staticmethod
    def canBeThreaded():
        """Return whether the servlet can be multithreaded.

        This value should not change during the lifetime of the object.
        The default implementation returns False.
        Note: This is not currently used.
        """
        return False

    @staticmethod
    def canBeReused():
        """Returns whether a single servlet instance can be reused.

        The default is True, but subclasses con override to return False.
        Keep in mind that performance may seriously be degraded if instances
        can't be reused. Also, there's no known good reasons not to reuse
        an instance. Remember the awake() and sleep() methods are invoked
        for every transaction. But just in case, your servlet can refuse
        to be reused.
        """
        return True

    # endregion Abilities

    # region Server side filesystem

    def serverSidePath(self, path=None):
        """Return the filesystem path of the page on the server."""
        if self._serverSidePath is None:
            self._serverSidePath = self._transaction.request().serverSidePath()
        if path:
            if path.startswith('/'):
                path = path[1:]
            return os.path.normpath(os.path.join(
                os.path.dirname(self._serverSidePath), path))
        return self._serverSidePath

    # endregion Server side filesystem

    # region Private

    def open(self):
        self._busy = True

    def close(self):
        if self._busy and self._factory:
            self._busy = False
            self._factory.returnServlet(self)

    def setFactory(self, factory):
        self._factory = factory

    # endregion Private



"""Generate Python code from PSP templates.

This module holds the classes that generate the Python code resulting
from the PSP template file. As the parser encounters PSP elements,
it creates a new Generator object for that type of element.
Each of these elements is put into a list maintained by the
ParseEventHandler object. When it comes time to output the source code,
each generator is called in turn to create its source.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

import os

from . import BraceConverter, PSPUtils

# This is global so that the ParseEventHandler and this module agree:
ResponseObject = 'res'


class GenericGenerator:
    """Base class for all the generators"""

    def __init__(self, ctxt=None):
        self._ctxt = ctxt
        self.phase = 'Service'


# pylint: disable=unused-argument

class ExpressionGenerator(GenericGenerator):
    """This class handles expression blocks.

    It simply outputs the (hopefully) python expression within the block
    wrapped with a _formatter() call.
    """

    def __init__(self, chars):
        self.chars = chars
        GenericGenerator.__init__(self)

    def generate(self, writer, phase=None):
        writer.println(
            f'res.write(_formatter({PSPUtils.removeQuotes(self.chars)}))')


class CharDataGenerator(GenericGenerator):
    """This class handles standard character output, mostly HTML.

    It just dumps it out. Need to handle all the escaping of characters.
    It's just skipped for now.
    """

    def __init__(self, chars):
        GenericGenerator.__init__(self)
        self.chars = chars

    def generate(self, writer, phase=None):
        # Quote any existing backslash so generated Python will not
        # interpret it when running.
        self.chars = self.chars.replace('\\', r'\\')
        # Quote any single quotes so it does not get confused with
        # our triple-quotes:
        self.chars = self.chars.replace('"', r'\"')
        self.generateChunk(writer)

    def generateChunk(self, writer, start=0, stop=None):
        writer.printIndent()  # gives one level of indentation
        writer.printChars(ResponseObject + '.write("""')
        writer.printChars(self.chars)
        writer.printChars('""")')
        writer.printChars('\n')

    def mergeData(self, cdGen):
        self.chars += cdGen.chars


class ScriptGenerator(GenericGenerator):
    """Generate scripts."""

    def __init__(self, chars, attrs):
        GenericGenerator.__init__(self)
        self.chars = chars

    def generate(self, writer, phase=None):
        self.chars = PSPUtils.normalizeIndentation(self.chars)
        if writer._useBraces:
            # Send lines to be output by the braces generator:
            braceConverter = BraceConverter.BraceConverter()
            lines = PSPUtils.splitLines(PSPUtils.removeQuotes(self.chars))
            for line in lines:
                braceConverter.parseLine(line, writer)
            return
        # Check for whitespace at beginning and if less than 2 spaces, remove:
        if self.chars.startswith(' ') and not self.chars.startswith('  '):
            self.chars = self.chars.lstrip()
        lines = PSPUtils.splitLines(PSPUtils.removeQuotes(self.chars))
        if not lines:
            return  # ignore any empty tag
        # userIndent check
        if lines[-1].endswith('$'):
            lastLine = lines[-1] = lines[-1][:-1]
            if not lastLine:
                lastLine = lines[-2]  # handle endscript marker on its own line
            count = 0
            while lastLine[count].isspace():
                count += 1
            userIndent = lastLine[:count]
        else:
            userIndent = writer._emptyString
            lastLine = lines[-1]
        # Print out code (moved from above):
        writer._userIndent = writer._emptyString  # reset to none
        writer.printList(lines)
        writer.printChars('\n')
        # Check for a block:
        # (Known issue: this fails if '#' part of string)
        commentStart = lastLine.find('#')
        if commentStart >= 0:
            lastLine = lastLine[:commentStart]
        blockCheck = lastLine.rstrip()
        if blockCheck.endswith(':'):
            writer.pushIndent()
            writer.println()
            writer._blockCount += 1
            # Check for end of block, "pass" by itself:
        if self.chars.strip() == 'pass' and writer._blockCount > 0:
            writer.popIndent()
            writer.println()
            writer._blockCount -= 1
        # Set userIndent for subsequent HTML:
        writer._userIndent = userIndent


class EndBlockGenerator(GenericGenerator):

    def __init__(self):
        GenericGenerator.__init__(self)

    def generate(self, writer, phase=None):
        if writer._blockCount > 0:
            writer.popIndent()
            writer.println()
            writer._blockCount -= 1
        writer._userIndent = writer._emptyString


class ScriptFileGenerator(GenericGenerator):
    """Add Python code at the file/module level."""

    def __init__(self, chars, attrs):
        GenericGenerator.__init__(self)
        self.phase = 'psp:file'
        self.attrs = attrs
        self.chars = chars

    def generate(self, writer, phase=None):
        writer.println('\n# File level user code')
        pySrc = PSPUtils.normalizeIndentation(self.chars)
        pySrc = PSPUtils.splitLines(PSPUtils.removeQuotes(pySrc))
        writer.printList(pySrc)


class ScriptClassGenerator(GenericGenerator):
    """Add Python code at the class level."""

    def __init__(self, chars, attrs):
        GenericGenerator.__init__(self)
        self.phase = 'psp:class'
        self.attrs = attrs
        self.chars = chars

    def generate(self, writer, phase=None):
        writer.println('# Class level user code\n')
        pySrc = PSPUtils.normalizeIndentation(self.chars)
        pySrc = PSPUtils.splitLines(PSPUtils.removeQuotes(pySrc))
        writer.printList(pySrc)


class MethodGenerator(GenericGenerator):
    """Generate class methods defined in the PSP page.

    There are two parts to method generation.
    This class handles getting the method name and parameters set up.
    """

    def __init__(self, chars, attrs):
        GenericGenerator.__init__(self)
        self.phase = 'Declarations'
        self.attrs = attrs

    def generate(self, writer, phase=None):
        writer.printIndent()
        writer.printChars('def ')
        writer.printChars(self.attrs['name'])
        writer.printChars('(')
        writer.printChars('self')
        if 'params' in self.attrs and self.attrs['params']:
            writer.printChars(', ')
            writer.printChars(self.attrs['params'])
        writer.printChars('):\n')
        if self.attrs['name'] == 'awake':
            writer._awakeCreated = True
            writer.pushIndent()
            writer.println('self.initPSP()\n')
            writer.popIndent()
            writer.println()


class MethodEndGenerator(GenericGenerator):
    """Part of class method generation.

    After MethodGenerator, MethodEndGenerator actually generates
    the code for the method body.
    """

    def __init__(self, chars, attrs):
        GenericGenerator.__init__(self)
        self.phase = 'Declarations'
        self.attrs = attrs
        self.chars = chars

    def generate(self, writer, phase=None):
        writer.pushIndent()
        writer.printList(PSPUtils.splitLines(
            PSPUtils.removeQuotes(self.chars)))
        writer.printChars('\n')
        writer.popIndent()


class IncludeGenerator(GenericGenerator):
    """Handle psp:include directives.

    This is a new version of this directive that actually
    forwards the request to the specified page.
    """

    _theFunction = (
        '__pspincludepath = "{}"\n'
        'self.transaction().application().includeURL('
        'self.transaction(), __pspincludepath)')

    def __init__(self, attrs, param, ctxt):
        GenericGenerator.__init__(self, ctxt)
        self.attrs = attrs
        self.param = param
        self.scriptGenerator = None

        self.url = attrs.get('path')
        if self.url is None:
            raise AttributeError('No path attribute in include')

        self.scriptGenerator = ScriptGenerator(
            self._theFunction.format(self.url), None)

    def generate(self, writer, phase=None):
        """Just insert theFunction."""
        self.scriptGenerator.generate(writer, phase)


class InsertGenerator(GenericGenerator):
    """Include files designated by the psp:insert syntax.

    If the attribute 'static' is set to True or 1, we include the file now,
    at compile time. Otherwise, we use a function added to every PSP page
    named ``__includeFile``, which reads the file at run time.
    """

    def __init__(self, attrs, param, ctxt):
        GenericGenerator.__init__(self, ctxt)
        self.attrs = attrs
        self.param = param
        self.scriptGenerator = None

        self.page = attrs.get('file')
        if not self.page:
            raise AttributeError('No file attribute in include')
        path = self._ctxt.resolveRelativeURI(self.page)
        if not os.path.exists(path):
            print(self.page)
            raise IOError(f'Invalid included file {path!r}')
        self.page = path

        self.static = str(attrs.get('static')).lower() in ('true', 'yes', '1')
        if not self.static:
            path = path.replace('\\', '\\\\')
            self.scriptGenerator = ScriptGenerator(
                f"self.__includeFile({path!r})", None)

    def generate(self, writer, phase=None):
        # JSP does this in the servlet. I'm doing it here because
        # I have triple quotes. Note: res.write statements inflate
        # the size of the resulting class file when it is cached.
        # Cut down on those by using a single res.write on the whole
        # file, after escaping any triple-double quotes.
        if self.static:
            with open(self.page) as f:
                data = f.read()
            data = data.replace('"""', r'\"""')
            writer.println(f'res.write("""{data}""")')
            writer.println()
        else:
            self.scriptGenerator.generate(writer, phase)



"""This module defines a class for writing responses using WSGI."""


class InvalidCommandSequence(ConnectionError):
    """Invalid command sequence error"""


class WSGIStreamOut:
    """This is a response stream to the client using WSGI.

    The key attributes of this class are:

    `_startResponse`:
       The start_response() function that is part of the WSGI protocol.
    `_autoCommit`:
        If True, the stream will automatically start sending data
        once it has accumulated `_bufferSize` data. This means that
        it will ask the response to commit itself, without developer
        interaction. By default, this is set to False.
    `_bufferSize`:
        The size of the data buffer. This is only used when autocommit
        is True. If not using autocommit, the whole response is
        buffered and sent in one shot when the servlet is done.
    `_useWrite`:
        Whether the write callable that is returned by start_response()
        shall be used to deliver the response.
    `flush()`:
        Send the accumulated response data now. Will ask the `Response`
        to commit if it hasn't already done so.
    """

    def __init__(self, startResponse,
                 autoCommit=False, bufferSize=8192,
                 useWrite=True, encoding='utf-8'):
        self._startResponse = startResponse
        self._autoCommit = autoCommit
        self._bufferSize = bufferSize
        self._useWrite = useWrite
        self._encoding = encoding
        self._committed = False
        self._needCommit = False
        self._buffer = b''
        self._chunks = []
        self._chunkLen = 0
        self._closed = False
        self._write = None
        self._iterable = []

    def startResponse(self, status, headers):
        """Start the response with the given status and headers."""
        if self._write or self._committed or self._closed:
            raise InvalidCommandSequence
        try:
            write = self._startResponse(status, headers)
            if self._useWrite:
                if write:
                    self._write = write
                else:
                    self._useWrite = False
                    raise ValueError(
                        'WSGI server does not support write().'
                        ' Try the setting WSGIWrite=False.')
        except Exception as e:
            print("Response Start Error:", e)
            raise ConnectionAbortedError from e

    def autoCommit(self):
        """Get the auto commit mode."""
        return self._autoCommit

    def setAutoCommit(self, autoCommit=True):
        """Set the auto commit mode."""
        self._autoCommit = bool(autoCommit)

    def bufferSize(self):
        """Get the buffer size."""
        return self._bufferSize

    def setBufferSize(self, bufferSize=8192):
        """Set the buffer size."""
        self._bufferSize = int(bufferSize)

    def flush(self):
        """Flush stream."""
        if self._closed:
            raise ConnectionAbortedError
        if not self._committed:
            if self._autoCommit:
                self._needCommit = True
            return
        if self._useWrite:
            if not self._write:
                raise InvalidCommandSequence
            write = self._write
        else:
            write = self._iterable.append
        try:
            self._buffer += b''.join(self._chunks)
        finally:
            self._chunks = []
            self._chunkLen = 0
        sent = 0
        buffer = self._buffer
        resLen = len(buffer)
        bufferSize = self._bufferSize
        while sent < resLen:
            try:
                write(buffer[sent:sent + bufferSize])
            except Exception as e:
                print("StreamOut Error:", e)
                self._closed = True
                raise ConnectionAbortedError from e
            sent += bufferSize
        self.pop(sent)

    def buffer(self):
        """Return accumulated data which has not yet been flushed.

        We want to be able to get at this data without having to call
        flush() first, so that we can (for example) integrate automatic
        HTML validation.
        """
        if self._buffer:  # if flush has been called, return what was flushed
            return self._buffer
        # otherwise return the buffered chunks
        return b''.join(self._chunks)

    def clear(self):
        """Try to clear any accumulated response data.

        Will fail if the response is already committed.
        """
        if self._committed:
            raise InvalidCommandSequence
        self._buffer = b''
        self._chunks = []
        self._chunkLen = 0

    def close(self):
        """Close this buffer. No more data may be sent."""
        self.flush()
        self._closed = True
        self._committed = True
        self._autoCommit = True
        self._write = None

    def closed(self):
        """Check whether we are closed to new data."""
        return self._closed

    def size(self):
        """Return the current size of the data held here."""
        return self._chunkLen + len(self._buffer)

    def prepend(self, output):
        """Add the output to the front of the response buffer.

        The output may be a byte string or a anything that can be converted
        to a string and encoded to a byte string using the output encoding.

        Invalid if we are already committed.
        """
        if self._committed or self._closed:
            raise InvalidCommandSequence
        if not isinstance(output, bytes):
            if not isinstance(output, str):
                output = str(output)
            output = output.encode(self._encoding)
        if self._buffer:
            self._buffer = output + self._buffer
        else:
            self._chunks.insert(0, output)
            self._chunkLen += len(output)

    def pop(self, count):
        """Remove count bytes from the front of the buffer."""
        self._buffer = self._buffer[count:]

    def committed(self):
        """Check whether the outptu is already committed"""
        return self._committed

    def needCommit(self):
        """Request for commitment.

        Called by the `HTTPResponse` instance that is using this instance
        to ask if the response needs to be prepared to be delivered.
        The response should then commit its headers, etc.
        """
        return self._needCommit

    def commit(self, autoCommit=True):
        """Called by the Response to tell us to go.

        If `_autoCommit` is True, then we will be placed into autoCommit mode.
        """
        self._committed = True
        self._autoCommit = autoCommit
        self.flush()

    def iterable(self):
        """Return the WSGI iterable."""
        return self._iterable

    def write(self, output):
        """Write output to the buffer.

        The output may be a byte string or a anything that can be converted
        to a string and encoded to a byte string using the output encoding.
        """
        if self._closed:
            raise ConnectionAbortedError
        if not isinstance(output, bytes):
            if not isinstance(output, str):
                output = str(output)
            output = output.encode(self._encoding)
        self._chunks.append(output)
        self._chunkLen += len(output)
        if self._autoCommit and self._chunkLen > self._bufferSize:
            self.flush()



"""A simple little module that organizes the actual page generation.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

from .StreamReader import StreamReader
from .ServletWriter import ServletWriter
from .PSPParser import PSPParser
from .ParseEventHandler import ParseEventHandler


class Compiler:
    """The main compilation class."""

    def __init__(self, context):
        self._ctxt = context

    def compile(self):
        """Compile the PSP context and return a set of all source files."""
        reader = StreamReader(self._ctxt.getPspFileName(), self._ctxt)
        reader.init()
        writer = ServletWriter(self._ctxt)
        self._ctxt.setPSPReader(reader)
        self._ctxt.setServletWriter(writer)
        parser = PSPParser(self._ctxt)
        handler = ParseEventHandler(self._ctxt, parser)
        parser.setEventHandler(handler)
        handler.beginProcessing()
        parser.parse()
        sourceFiles = set(reader.sourceFiles)
        handler.endProcessing()
        writer.close()
        return sourceFiles



import unittest

from TaskKit.Scheduler import Scheduler


class TaskKitTest(unittest.TestCase):

    def setUp(self):
        self._scheduler = Scheduler()

    def testSchedulerStarts(self):
        scheduler = self._scheduler
        scheduler.start()

    def tearDown(self):
        self._scheduler.stop()
        self._scheduler = None



import unittest

from io import StringIO

from WebUtils.HTMLTag import HTMLReader


class HTMLTagTest(unittest.TestCase):

    def setUp(self):
        self._html = """\
<html>
<head>
    <title>Example</title>
</head>
<body lang="en">
    <p>What's up, <em>doc</em>?</p>
    <hr>
    <table id="dataTable">
        <tr> <th> x </th> <th> y </th> </tr>
        <tr><td class="datum">0</td><td class="datum">0</td></tr>
    </table>
</body>
</html>"""

    def tearDown(self):
        del self._html

    def testBasics(self):
        reader = HTMLReader()
        tag = reader.readString('<html> </html>')
        self.assertEqual(tag.name(), 'html')
        self.assertEqual(reader.rootTag(), tag)
        self.assertTrue(reader.filename() is None)
        out = StringIO()
        tag.pprint(out)
        self.assertEqual(out.getvalue(), '<html>\n</html>\n')

    def testReuseReader(self):
        reader = HTMLReader()
        reader.readString('<html> </html>')
        tag = reader.readString('<html> <body> </body> </html>')
        self.assertFalse(reader.rootTag() is None)
        self.assertEqual(reader.rootTag(), tag)

        tag = reader.readString('<html> </html>', retainRootTag=0)
        self.assertFalse(tag is None)
        self.assertTrue(reader.rootTag() is None)

    def testAccess(self):
        html = HTMLReader().readString(self._html)

        # Name
        self.assertEqual(html.name(), 'html')

        # Attrs
        self.assertEqual(html.numAttrs(), 0)
        self.assertFalse(html.hasAttr('foo'))
        self.assertRaises(KeyError, html.attr, 'foo')
        self.assertTrue(html.attr('foo', None) is None)

        # Children and subtags, when both are the same.
        for numFoos, fooAt, foos in [
                [html.numChildren, html.childAt, html.children],
                [html.numSubtags, html.subtagAt, html.subtags]]:
            self.assertEqual(numFoos(), 2)
            self.assertEqual(len(foos()), 2)
            self.assertEqual(fooAt(0).name(), 'head')
            self.assertEqual(fooAt(1).name(), 'body')

        # Children and subtags when they're different
        body = html.subtagAt(1)
        p = body.subtagAt(0)
        self.assertEqual(p.name(), 'p')
        self.assertEqual(p.numChildren(), 3)
        self.assertEqual(p.numSubtags(), 1)

    def testMatchingAttr(self):
        html = HTMLReader().readString(self._html)
        self.assertEqual(
            html.tagWithMatchingAttr('lang', 'en').name(), 'body')
        self.assertEqual(
            html.tagWithMatchingAttr('id', 'dataTable').name(), 'table')
        self.assertEqual(html.tagWithId('dataTable').name(), 'table')

    def testInvalidHTML(self):
        from WebUtils.HTMLTag import (
            HTMLTagUnbalancedError, HTMLTagIncompleteError)

        reader = HTMLReader()

        html = '<html> <body> <table> </body> </html>'
        self.assertRaises(HTMLTagUnbalancedError, reader.readString, html)

        html = '<html> <body>'
        self.assertRaises(HTMLTagIncompleteError, reader.readString, html)



import pkg_resources

from PlugIn import PlugIn


class PlugInLoader:

    def __init__(self, app):
        self.app = app

    def loadPlugIn(self, name, module, verbose=True):
        """Load and return the given plug-in.

        May return None if loading was unsuccessful (in which case this method
        prints a message saying so). Used by `loadPlugIns` (note the **s**).
        """
        try:
            plugIn = PlugIn(self.app, name, module)
            willNotLoadReason = plugIn.load(verbose=verbose)
            if willNotLoadReason:
                print(f'    Plug-in {name} cannot be loaded because:\n'
                      f'    {willNotLoadReason}')
                return None
            plugIn.install()
        except Exception:
            print()
            print(f'Plug-in {name} raised exception.')
            raise
        return plugIn

    def loadPlugIns(self, plugInNames=None, verbose=None):
        """Load all plug-ins.

        A plug-in allows you to extend the functionality of Webware without
        necessarily having to modify its source. Plug-ins are loaded by
        Application at startup time, just before listening for requests.
        See the docs in `PlugIn` for more info.
        """
        if plugInNames is None:
            plugInNames = self.app.setting('PlugIns')
        if verbose is None:
            verbose = self.app.setting('PrintPlugIns')

        if verbose:
            print('Plug-ins list:', ', '.join(plugInNames))

        entryPoints = {
            entry_point.name: entry_point for entry_point
            in pkg_resources.iter_entry_points('webware.plugins')}

        plugIns = {}
        for name in plugInNames:
            if name in plugIns:
                if verbose:
                    print(f'Plug-in {name} has already been loaded.')
                    continue
            entry_point = entryPoints.get(name)
            if not entry_point:
                if verbose:
                    print(f'Plug-in {name} has not entry point.')
                    continue
            module = entry_point.load()
            plugIn = self.loadPlugIn(name, module, verbose=verbose)
            if plugIn:
                plugIns[name] = plugIn

        if verbose:
            print()

        return plugIns



"""Mock redis-py module."""

import sys
from copy import copy

sys.modules['redis'] = sys.modules[__name__]

data = {}  # our mock redis


class ConnectionPool:

    def __init__(self, redis):
        self._redis = redis

    def disconnect(self):
        self._redis._connected = False


# pylint: disable=unused-argument, invalid-name

class StrictRedis:
    """Mock Redis client."""

    def __init__(self, host='localhost', port=6379, db=0, password=None):
        self.connection_pool = ConnectionPool(self)
        self._connected = True

    def setex(self, name, time, value):
        if self._connected:
            if value is not None:
                data[name] = value

    def get(self, name):
        if self._connected:
            return copy(data.get(name))

    def delete(self, *names):
        if self._connected:
            for name in names:
                del data[name]

    def exists(self, name):
        if self._connected:
            return name in data

    def keys(self, pattern='*'):
        if self._connected:
            if not pattern.endswith('*'):
                raise ValueError('bad pattern')
            pattern = pattern[:-1]
            if not pattern:
                return list(data)
            return [k for k in data if k.startswith(pattern)]

    def flushdb(self):
        if self._connected:
            data.clear()



from math import sin, pi

from io import BytesIO

from MiscUtils.Funcs import asclocaltime
from .ExamplePage import ExamplePage

try:
    from PIL import Image as pilImage, ImageDraw, ImageFont
except ImportError:
    pilImage = ImageDraw = ImageFont = None


w, h = 500, 200  # the image size


def t(p):
    """Map coordinates: (x=0..2pi, y=-1.25..1.25) => (0..X, Y..0)"""
    return int(.5 * w * p[0] / pi + .5), int(.4 * h * (1.25 - p[1]) + .5)


colors = (255, 255, 255), (0, 0, 0), (0, 0, 255), (255, 0, 0)
white, black, blue, red = range(4)


class Drawing:
    """Simple wrapper class for drawing the example image."""

    def __init__(self):
        self._image = pilImage.new('RGB', (w, h), colors[white])
        self._draw = ImageDraw.Draw(self._image)
        for font in 'Tahoma Verdana Arial Helvetica'.split():
            try:
                font = ImageFont.truetype(font + '.ttf', 12)
            except (AttributeError, IOError):
                font = None
            if font:
                break
        else:
            try:
                font = ImageFont.load_default()
            except (AttributeError, IOError):
                font = None
        self._font = font

    def text(self, pos, string, color):
        color = colors[color]
        self._draw.text(t(pos), string, color, font=self._font)

    def lines(self, points, color):
        color = colors[color]
        self._draw.line(list(map(t, points)), color)

    def png(self):
        s = BytesIO()
        self._image.save(s, 'png')
        return s.getvalue()


class ImageDemo(ExamplePage):
    """Dynamic image generation example.

    This example creates an image of a sinusoid.

    For more information on generating graphics, see
    https://wiki.python.org/moin/GraphicsAndImages.

    This example uses the Python Imaging Library (Pillow).
    """

    saveToFile = False

    def defaultAction(self):
        fmt = self.request().field('fmt', None)
        if fmt == '.png' and pilImage:
            image = self.generatePNGImage()
            res = self.response()
            res.setHeader("Content-Type", "image/png")
            res.setHeader("Content-Length", str(len(image)))
            if self.saveToFile:
                res.setHeader("Content-Disposition",
                              "attachment; filename=demo.png")
            self.write(image)
        else:
            self.writeHTML()

    def writeContent(self):
        wr = self.writeln
        wr('<h2>Webware Image Generation Demo</h2>')
        lib = 'Python Imaging Library (Pillow)'
        if pilImage:
            wr('<img src="ImageDemo?fmt=.png" alt="Generated example image"'
               f' width="{w:d}" height="{h:d}">')
            wr(f'<p>This image has just been generated using the {lib}.</p>')
        else:
            wr('<h4 style="color:red">Sorry: No imaging tool available.</h4>')
            src = 'https://pypi.org/project/Pillow/'
            wr(f'<p>This example requires the <a href="{src}">{lib}</a>.</p>')

    def generatePNGImage(self):
        """Generate and return a PNG example image."""
        def f(x):
            return x, sin(x)
        draw = Drawing()
        draw.text((2.7, 0.8), 'y=sin(x)', black)
        draw.text((0.2, -0.8), f'created: {asclocaltime()}', red)
        draw.lines(((0, 0), (2 * pi, 0)), black)  # x-axis
        draw.lines(((0, -1), (0, 1)), black)  # y-axis
        draw.lines([f(x * 2 * pi / w) for x in range(w + 1)], blue)
        return draw.png()



import os

from SidebarPage import SidebarPage


class ExamplePage(SidebarPage):
    """Webware page template class for displaying examples.

    The convention for picking up examples from Webware plug-ins is:
      * Properties.py must have at least this:
            webwareConfig = {
                'examplePages': []
            }
        But usually specifies various pages:
            webwareConfig = {
                'examplesPages': [
                    'foo',
                    'bar',
                ]
            }
      * The plug-in must have an Examples/ directory.
      * That directory must have an index.* or Main.* which
        inherits from ExamplePage.
      * The implementation can pass in which case a menu of
        the pages for that plug-in is written:
            # Main.py
            from .Examples.ExamplePage import ExamplePage
            class Main(ExamplePage):
                pass

    If webwareConfig['examplesPages'] is non-existent or None, then
    no examples will be available for the plug-in.

    If the Webware Examples context is not present in the first place,
    then there is no access to the plug-in examples.
    """

    def cornerTitle(self):
        return 'Webware Examples'

    def isDebugging(self):
        return False

    def examplePages(self, plugInName=None):
        """Get a list of all example pages.

        Returns a list of all the example pages for our particular plug-in.
        These can be used in the sidebar or in the main content area to
        give easy access to the other example pages.
        """
        if plugInName is None or plugInName == 'Examples':
            # Special case: We're in Webware examples
            from Properties import webwareConfig
            return webwareConfig['examplePages']
        # Any other plug-in:
        try:
            plugIn = self.application().plugIn(plugInName)
        except KeyError:
            return []
        else:
            return plugIn.examplePages()

    def writeSidebar(self):
        self.writeExamplesMenu()
        self.writeOtherMenu()
        self.writeWebwareSidebarSections()

    def writeExamplesMenu(self):
        servletPath = self.request().servletPath()
        self.menuHeading('Examples')
        ctx = self.request().contextName().split('/', 2)
        plugIns = self.application().plugIns()
        plugInName = len(ctx) > 1 and ctx[1] == 'Examples' and ctx[0]
        # Webware
        self.menuItem('Webware', f'{servletPath}/Examples/')
        if not plugInName:
            self.writeExamplePagesItems()
        # Other plug-ins
        for name, plugIn in plugIns.items():
            if plugIn.hasExamplePages():
                link = f'{servletPath}/{plugIn.examplePagesContext()}/'
                self.menuItem(name, link)
                if name == plugInName:
                    self.writeExamplePagesItems(name)

    def writeExamplePagesItems(self, pluginName=None):
        for page in self.examplePages(pluginName):
            self.menuItem(page, page, indentLevel=2)

    def writeOtherMenu(self):
        self.menuHeading('Other')
        self.menuItem(
            'View source of<br>' + self.title(),
            self.request().uriWebwareRoot() + 'Examples/View?filename=' +
            os.path.basename(self.request().serverSidePath()))

    def writeContent(self):
        wr = self.writeln
        wr('<div style="padding-left:22em"><table>')
        for page in self.examplePages(
                self.request().contextName().split('/', 1)[0]):
            wr('<tr><td style="background-color:#e8e8f0">'
               f'<a href={page}>{page}</a></td></tr>')
        wr('</table></div>')



#!/usr/bin/env python3

"""The Application singleton.

`Application` is the main class that sets up and dispatches requests.
This is done using the WSGI protocol, so an `AppServer` class is not
needed and not contained in Webware for Python anymore.
`Application` receives the input via WSGI and turns it into `Transaction`,
`HTTPRequest`, `HTTPResponse`, and `Session`.

Settings for Application are taken from ``Configs/Application.config``,
which is used for many global settings, even if they aren't closely tied
to the Application object itself.
"""

import atexit
import os
import signal
import sys

from time import time, localtime

from MiscUtils import NoDefault
from MiscUtils.Funcs import asclocaltime
from MiscUtils.NamedValueAccess import valueForName
from TaskKit.Scheduler import Scheduler
from WebUtils.Funcs import requestURI

from ConfigurableForServerSidePath import ConfigurableForServerSidePath
from ImportManager import ImportManager
from ExceptionHandler import ExceptionHandler
from HTTPRequest import HTTPRequest
from HTTPExceptions import HTTPException, HTTPSessionExpired
from Transaction import Transaction
from WSGIStreamOut import WSGIStreamOut
from PlugInLoader import PlugInLoader

import URLParser

debug = False

defaultConfig = dict(
    ActivityLogFilename='Logs/Activity.csv',
    ActivityLogColumns=[
        'request.remoteAddress', 'request.method',
        'request.uri', 'response.size',
        'servlet.name', 'request.timeStamp',
        'transaction.duration',
        'transaction.errorOccurred'
    ],
    AlwaysSaveSessions=True,
    AppLogFilename='Logs/Application.log',
    CacheDir='Cache',
    CacheServletClasses=True,
    CacheServletInstances=True,
    CheckInterval=None,
    Contexts={
        'default': 'Examples',
        'Admin': 'Admin',
        'Examples': 'Examples',
        'Testing': 'Testing',
    },
    Debug=dict(
        Sessions=False,
    ),
    DirectoryFile=['index', 'Index', 'main', 'Main'],
    EnterDebuggerOnException=False,
    EmailErrors=False,
    EmailErrorReportAsAttachment=False,
    ErrorEmailServer='localhost',
    ErrorEmailHeaders={
        'From': 'webware@mydomain',
        'To': ['webware@mydomain'],
        'Reply-To': 'webware@mydomain',
        'Content-Type': 'text/html',
        'Subject': 'Error'
    },
    ErrorLogFilename='Logs/Errors.csv',
    ErrorMessagesDir='ErrorMsgs',
    ErrorPage=None,
    ExtensionCascadeOrder=['.py', '.psp', '.html'],
    ExtensionsToIgnore={
        '.pyc', '.pyo', '.tmpl', '.bak', '.py_bak',
        '.py~', '.psp~', '.html~', '.tmpl~'
    },
    ExtensionsToServe=[],
    ExtraPathInfo=True,
    FancyTracebackContext=5,
    FilesToHide={
        '.*', '*~', '*.bak', '*.py_bak', '*.tmpl',
        '*.pyc', '*.pyo', '__init__.*', '*.config'
    },
    FilesToServe=[],
    IgnoreInvalidSession=True,
    IncludeEditLink=True,
    IncludeFancyTraceback=False,
    LogActivity=True,
    LogErrors=True,
    MaxValueLengthInExceptionReport=500,
    OutputEncoding='utf-8',
    PlugIns=['MiscUtils', 'WebUtils', 'TaskKit', 'UserKit', 'PSP'],
    PrintConfigAtStartUp=True,
    PrintPlugIns=True,
    ReloadServletClasses=False,
    ReportRPCExceptionsInWebware=True,
    ResponseBufferSize=8 * 1024,  # 8 kBytes
    RetainSessions=True,
    RPCExceptionReturn='traceback',
    RunTasks=True,
    SaveErrorMessages=True,
    SecureSessionCookie=True,
    SessionCookiePath=None,
    SessionModule='Session',
    SessionName='_SID_',
    SessionPrefix='',
    SessionStore='Dynamic',
    SessionStoreDir='Sessions',
    SessionTimeout=60,
    ShowDebugInfoOnErrors=False,
    SilentURIs=None,
    UnknownFileTypes=dict(
        ReuseServlets=True,
        Technique='serveContent',  # or redirectSansScript
        CacheContent=False,
        MaxCacheContentSize=128 * 1024,
        ReadBufferSize=32 * 1024
    ),
    UseAutomaticPathSessions=False,
    UseCascadingExtensions=True,
    UseCookieSessions=True,
    UserErrorMessage=(
        'The site is having technical difficulties with this page.'
        ' An error has been logged, and the problem will be fixed'
        ' as soon as possible. Sorry!'
    ),
    UseSessionSweeper=True,
    Verbose=True,
    WSGIWrite=True,  # use write callable with WSGI
)


class EndResponse(Exception):
    """End response exception.

    Used to prematurely break out of the `awake()`/`respond()`/`sleep()`
    cycle without reporting a traceback. During servlet processing,
    if this exception is caught during `awake()` or `respond()` then `sleep()`
    is called and the response is sent. If caught during `sleep()`,
    processing ends and the response is sent.
    """


class Application(ConfigurableForServerSidePath):
    """The Application singleton.

    Purpose and usage are explained in the module docstring.
    """

    # region Init

    def __init__(self, path=None, settings=None, development=None):
        """Sets up the Application.

        You can specify the path of the application working directory,
        a dictionary of settings to override in the configuration,
        and whether the application should run in development mode.
        """
        ConfigurableForServerSidePath.__init__(self)
        if path is None:
            path = os.getcwd()
        self._serverSidePath = os.path.abspath(path)
        self._webwarePath = os.path.abspath(os.path.dirname(__file__))

        if not os.path.isfile(self.configFilename()):
            print("ERROR: The application cannot be started:")
            print(f"Configuration file {self.configFilename()} not found.")
            raise RuntimeError('Configuration file not found')

        if development is None:
            development = bool(os.environ.get('WEBWARE_DEVELOPMENT'))
        self._development = development

        self.initVersions()

        self._shutDownHandlers = []
        self._plugIns = {}
        self._requestID = 0

        self._imp = ImportManager()

        appConfig = self.config()  # get and cache the configuration
        if settings:
            appConfig.update(settings)

        self._verbose = self.setting('Verbose')
        if self._verbose:
            self._silentURIs = self.setting('SilentURIs')
            if self._silentURIs:
                import re
                self._silentURIs = re.compile(self._silentURIs)
        else:
            self._silentURIs = None
        self._outputEncoding = self.setting('OutputEncoding')
        self._responseBufferSize = self.setting('ResponseBufferSize')
        self._wsgiWrite = self.setting('WSGIWrite')
        if self.setting('CheckInterval') is not None:
            sys.setswitchinterval(self.setting('CheckInterval'))

        logFilename = self.setting('AppLogFilename')
        if logFilename:
            sys.stderr = sys.stdout = open(logFilename, 'a', buffering=1)

        self.initErrorPage()
        self.printStartUpMessage()

        # Initialize task manager:
        if self.setting('RunTasks'):
            self._taskManager = Scheduler(
                daemon=True, exceptionHandler=self.handleException)
            self._taskManager.start()
        else:
            self._taskManager = None

        # Define this before initializing URLParser, so that contexts have a
        # chance to override this. Also be sure to define it before loading the
        # sessions, in case the loading of the sessions causes an exception.
        self._exceptionHandlerClass = ExceptionHandler

        self.makeDirs()
        self.initSessions()

        URLParser.initApp(self)
        self._rootURLParser = URLParser.ContextParser(self)

        self._startTime = time()

        if self.setting('UseSessionSweeper'):
            self.startSessionSweeper()

        self._plugInLoader = None
        self.loadPlugIns()

        self._needsShutDown = [True]
        atexit.register(self.shutDown)
        self._sigTerm = signal.signal(signal.SIGTERM, self.sigTerm)
        try:
            self._sigHup = signal.signal(signal.SIGHUP, self.sigTerm)
        except AttributeError:
            pass  # SIGHUP does not exist on Windows

    def initErrorPage(self):
        """Initialize the error page related attributes."""
        for path in (self._serverSidePath,
                     os.path.dirname(os.path.abspath(__file__))):
            error404file = os.path.join(path, 'error404.html')
            try:
                with open(error404file) as f:
                    self._error404 = f.read()
            except Exception:
                continue
            else:
                break
        else:
            self._error404 = None
        urls = self.setting('ErrorPage') or None
        if urls:
            try:
                errors = urls.keys()
            except AttributeError:
                errors = ['Exception']
                urls = {errors[0]: urls}
            for err in errors:
                if urls[err] and not urls[err].startswith('/'):
                    urls[err] = '/' + urls[err]
        self._errorPage = urls

    def initSessions(self):
        """Initialize all session related attributes."""
        setting = self.setting
        self._sessionPrefix = setting('SessionPrefix') or ''
        if self._sessionPrefix:
            if self._sessionPrefix == 'hostname':
                from MiscUtils.Funcs import hostName
                self._sessionPrefix = hostName()
            self._sessionPrefix += '-'
        self._sessionTimeout = setting('SessionTimeout') * 60
        self._sessionName = (setting('SessionName') or
                             self.defaultConfig()['SessionName'])
        self._autoPathSessions = setting('UseAutomaticPathSessions')
        self._alwaysSaveSessions = setting('AlwaysSaveSessions')
        self._retainSessions = setting('RetainSessions')
        moduleName = setting('SessionModule')
        className = moduleName.rpartition('.')[2]
        try:
            exec(f'from {moduleName} import {className}')
            cls = locals()[className]
            if not isinstance(cls, type):
                raise ImportError
            self._sessionClass = cls
        except ImportError:
            print(f"ERROR: Could not import Session class '{className}'"
                  f" from module '{moduleName}'")
            self._sessionClass = None
        moduleName = setting('SessionStore')
        if moduleName in (
                'Dynamic', 'File', 'Memcached', 'Memory', 'Redis', 'Shelve'):
            moduleName = f'Session{moduleName}Store'
        className = moduleName.rpartition('.')[2]
        try:
            exec(f'from {moduleName} import {className}')
            cls = locals()[className]
            if not isinstance(cls, type):
                raise ImportError
            self._sessions = cls(self)
        except ImportError as err:
            print(f"ERROR: Could not import SessionStore class '{className}'"
                  f" from module '{moduleName}':\n{err}")
            self._sessions = None

    def makeDirs(self):
        """Make sure some standard directories are always available."""
        setting = self.setting
        self._cacheDir = self.serverSidePath(
            setting('CacheDir') or 'Cache')
        self._errorMessagesDir = self.serverSidePath(
            setting('ErrorMessagesDir') or 'ErrorMsgs')
        self._sessionDir = self.serverSidePath(
            setting('SessionStoreDir') or 'Sessions')
        for path in (self.serverSidePath('Logs'), self._cacheDir,
                     self._errorMessagesDir, self._sessionDir):
            if path and not os.path.exists(path):
                os.makedirs(path)

    def initVersions(self):
        """Get and store versions.

        Initialize attributes that stores the Webware version as
        both tuple and string. These are stored in the Properties.py files.
        """
        from MiscUtils.PropertiesObject import PropertiesObject
        props = PropertiesObject(os.path.join(
            self.webwarePath(), 'Properties.py'))
        self._webwareVersion = props['version']
        self._webwareVersionString = props['versionString']
        if tuple(sys.version_info) < tuple(props['requiredPyVersion']):
            pythonVersion = '.'.join(map(str, sys.version_info))
            requiredVersion = props['requiredPyVersionString']
            raise RuntimeError(
                f' Required Python version is {requiredVersion},'
                f' but actual version is {pythonVersion}.')

    def taskManager(self):
        """Accessor: `TaskKit.Scheduler` instance."""
        return self._taskManager

    def startSessionSweeper(self):
        """Start session sweeper.

        Starts the session sweeper, `Tasks.SessionTask`, which deletes
        session objects (and disk copies of those objects) that have expired.
        """
        if self._sessionTimeout:
            tm = self.taskManager()
            if tm:
                from Tasks import SessionTask
                task = SessionTask.SessionTask(self._sessions)
                sweepInterval = self._sessionTimeout / 10
                tm.addPeriodicAction(time() + sweepInterval, sweepInterval,
                                     task, "SessionSweeper")
                print("Session sweeper has started.")

    def shutDown(self):
        """Shut down the application.

        Called when the interpreter is terminated.
        """
        try:  # atomic safety check
            self._needsShutDown.pop()
        except IndexError:  # shut down already initiated
            return
        print("Application is shutting down...")
        atexit.unregister(self.shutDown)
        if self._sessions:
            self._sessions.storeAllSessions()
        tm = self.taskManager()
        if tm:
            tm.stop()
        # Call all registered shutdown handlers
        for shutDownHandler in self._shutDownHandlers:
            try:
                shutDownHandler()
            except Exception:
                pass
        print("Application has been successfully shut down.")

    def addShutDownHandler(self, func):
        """Add a shutdown handler.

        Functions added through `addShutDownHandler` will be called when
        the Application is shutting down. You can use this hook to close
        database connections, clean up resources, save data to disk, etc.
        """
        self._shutDownHandlers.append(func)

    def sigTerm(self, _signum, _frame):
        """Signal handler for terminating the process."""
        if self._needsShutDown:
            print("\nApplication has been signaled to terminate.")
            self.shutDown()

    # endregion Init

    # region Config

    def defaultConfig(self):
        """The default Application.config."""
        return defaultConfig  # defined on the module level

    def configFilename(self):
        """The configuration file path."""
        return self.serverSidePath('Configs/Application.config')

    def configReplacementValues(self):
        """Get config values that need to be escaped."""
        return dict(
            ServerSidePath=self._serverSidePath,
            WebwarePath=self._webwarePath,
            Development=self._development)

    def development(self):
        """Whether the application shall run in development mode"""
        return self._development

    # endregion Config

    # region Versions

    def webwareVersion(self):
        """Return the Webware version as a tuple."""
        return self._webwareVersion

    def webwareVersionString(self):
        """Return the Webware version as a printable string."""
        return self._webwareVersionString

    # endregion Version

    # region Sessions

    def session(self, sessionId, default=NoDefault):
        """The session object for `sessionId`.

        Raises `KeyError` if session not found and no default is given.
        """
        if default is NoDefault:
            return self._sessions[sessionId]
        return self._sessions.get(sessionId, default)

    def hasSession(self, sessionId):
        """Check whether session `sessionId` exists."""
        return sessionId in self._sessions

    def sessions(self):
        """A dictionary of all the session objects."""
        return self._sessions

    def createSessionForTransaction(self, trans):
        """Get the session object for the transaction.

        If the session already exists, returns that, otherwise creates
        a new session.

        Finding the session ID is done in `Transaction.sessionId`.
        """
        debug = self.setting('Debug').get('Sessions')
        if debug:
            prefix = '>> [session] createSessionForTransaction:'
        sessId = trans.request().sessionId()
        if debug:
            print(prefix, 'sessId =', sessId)
        if sessId:
            try:
                session = self.session(sessId)
                if debug:
                    print(prefix, 'retrieved session =', session)
            except KeyError:
                trans.request().setSessionExpired(1)
                if not self.setting('IgnoreInvalidSession'):
                    raise HTTPSessionExpired from None
                sessId = None
        if not sessId:
            session = self._sessionClass(trans)
            self._sessions[session.identifier()] = session
            if debug:
                print(prefix, 'created session =', session)
        trans.setSession(session)
        return session

    def createSessionWithID(self, trans, sessionID):
        """Create a session object with our session ID."""
        session = self._sessionClass(trans, sessionID)
        # Replace the session if it didn't already exist,
        # otherwise we just throw it away.  setdefault is an atomic
        # operation so this guarantees that 2 different
        # copies of the session with the same ID never get placed into
        # the session store, even if multiple threads are calling
        # this method simultaneously.
        trans.application()._sessions.setdefault(sessionID, session)

    def sessionTimeout(self, _trans):
        """Get the timeout (in seconds) for a user session.

        Overwrite to make this transaction dependent.
        """
        return self._sessionTimeout

    def sessionPrefix(self, _trans):
        """Get the prefix string for the session ID.

        Overwrite to make this transaction dependent.
        """
        return self._sessionPrefix

    def sessionName(self, _trans):
        """Get the name of the field holding the session ID.

        Overwrite to make this transaction dependent.
        """
        return self._sessionName

    def sessionCookiePath(self, trans):
        """Get the cookie path for this transaction.

        If not path is specified in the configuration setting,
        the servlet path is used for security reasons, see:
        https://www.helpnetsecurity.com/2004/06/27/cookie-path-best-practice/
        """
        return self.setting('SessionCookiePath') or (
            trans.request().servletPath() + '/')

    # endregion Sessions

    # region Misc Access

    def serverSidePath(self, path=None):
        """Get the server-side path.

        Returns the absolute server-side path of the Webware application.
        If the optional path is passed in, then it is joined with the
        server side directory to form a path relative to the working directory.
        """
        if path:
            return os.path.normpath(
                os.path.join(self._serverSidePath, path))
        return self._serverSidePath

    def webwarePath(self):
        """Return the Webware path."""
        return self._webwarePath

    @staticmethod
    def name():
        """The name by which this was started. Usually `Application`."""
        return sys.argv[0]

    def outputEncoding(self):
        """Get the default output encoding of this application."""
        return self._outputEncoding

    # endregion Misc Access

    # region Activity Log

    def printStartUpMessage(self):
        """Print a little intro to the activity log."""
        print('Webware for Python', self.webwareVersionString(), 'Application')
        print()
        print('Process id:', os.getpid())
        print('Start date/time:', asclocaltime())
        print('Python:', sys.version)
        print()
        if self.development():
            print("Running in development mode")
            print()
        if self.setting('PrintConfigAtStartUp'):
            self.printConfig()

    def writeActivityLog(self, trans):
        """Write an entry to the activity log.

        Writes an entry to the script log file. Uses settings
        ``ActivityLogFilename`` and ``ActivityLogColumns``.
        """
        filename = self.serverSidePath(
            self.setting('ActivityLogFilename'))
        if os.path.exists(filename):
            f = open(filename, 'a')
        else:
            f = open(filename, 'w')
            f.write(','.join(self.setting('ActivityLogColumns')) + '\n')
        values = []
        objects = dict(
            application=self, transaction=trans,
            request=trans.request(), response=trans.response(),
            servlet=trans.servlet(),
            # don't cause creation of session here:
            session=trans._session)
        for column in self.setting('ActivityLogColumns'):
            try:
                value = valueForName(objects, column)
            except Exception:
                value = '(unknown)'
            if isinstance(value, float):
                # probably need more flexibility in the future
                value = f'{value:02f}'
            else:
                value = str(value)
            values.append(value)
        f.write(','.join(values) + '\n')
        f.close()

    def startTime(self):
        """Return the time the application was started.

        The time is given as seconds, like time().
        """
        return self._startTime

    def numRequests(self):
        """Return the number of requests.

        Returns the number of requests received by this application
        since it was launched.
        """
        return self._requestID

    # endregion Activity Log

    # region Request Dispatching

    # These are the entry points from WSGI, which take a raw request,
    # turn it into a transaction, run the transaction, and clean up.

    def dispatchRawRequest(self, requestDict, strmOut):
        """Dispatch a raw request.

        Dispatch a request as passed from the web server via WSGI.

        This method creates the request, response, and transaction objects,
        then runs (via `runTransaction`) the transaction. It also catches any
        exceptions, which are then passed on to `handleExceptionInTransaction`.
        """
        request = self.createRequestForDict(requestDict)
        if request:
            trans = Transaction(application=self, request=request)
            if trans:
                request.setTransaction(trans)
                response = request.responseClass()(trans, strmOut)
                if response:
                    trans.setResponse(response)
                    self.runTransaction(trans)
                    try:
                        trans.response().deliver()
                    except ConnectionAbortedError as err:
                        trans.setError(err)
                    response.clearTransaction()
                    # release possible servlets on the stack
                    while True:
                        servlet = request.pop()
                        if not servlet:
                            break
                        self.returnServlet(servlet)
                    # get current servlet (this may have changed)
                    servlet = trans.servlet()
                    if servlet:
                        # return the current servlet to its pool
                        self.returnServlet(servlet)
                if self.setting('LogActivity'):
                    self.writeActivityLog(trans)
            request.clearTransaction()
        return trans

    @staticmethod
    def createRequestForDict(requestDict):
        """Create request object for a given dictionary.

        Create a request object (subclass of `Request`) given the raw
        dictionary as passed by the web server via WSGI.

        The class of the request may be based on the contents of the
        dictionary (though only `HTTPRequest` is currently created),
        and the request will later determine the class of the response.

        Called by `dispatchRawRequest`.
        """
        fmt = requestDict['format']
        # Maybe an Email adapter would make a request with a format of Email,
        # and an EmailRequest would be generated. For now just CGI/HTTPRequest.
        if fmt != 'CGI':
            raise ValueError('Only CGI format request dicts are supported')
        request = HTTPRequest(requestDict)
        # The request object is stored for tracking/debugging purposes.
        requestDict['requestObject'] = request
        return request

    def runTransaction(self, trans):
        """Run transaction.

        Executes the transaction, handling HTTPException errors.
        Finds the servlet (using the root parser, probably
        `URLParser.ContextParser`, to find the servlet for the transaction,
        then calling `runTransactionViaServlet`.

        Called by `dispatchRawRequest`.
        """
        findServlet = self.rootURLParser().findServletForTransaction
        try:
            # remove the session identifier from the path
            self.removePathSession(trans)
            # determine the context and the servlet for the transaction
            servlet = findServlet(trans)
            # handle session field only now, because the name of the
            # session id field can depend on the context
            self.handlePathSession(trans)
            # now everything is set, run the transaction
            self.runTransactionViaServlet(servlet, trans)
        except EndResponse:
            pass
        except ConnectionAbortedError as err:
            trans.setError(err)
        except Exception as err:
            urls = {}
            while True:
                trans.setError(err)
                isHTTPException = isinstance(err, HTTPException)
                if isHTTPException:
                    err.setTransaction(trans)  # pylint: disable=no-member
                if trans.response().isCommitted():
                    # response already committed, cannot display error
                    isHTTPException = False
                    break
                if not self._errorPage:
                    # no custom error page configured
                    break
                # get custom error page for this exception
                errClass = err.__class__
                url = errClass and self.errorPage(errClass)
                if isHTTPException and not url:
                    # get custom error page for status code
                    code = err.code()  # pylint: disable=no-member
                    if code in self._errorPage:
                        url = self._errorPage[code]
                if not url or url in urls:
                    # If there is no custom error page configured,
                    # or we get into a circular chain of error pages,
                    # then we fall back to standard error handling.
                    break
                urls[url] = None
                # forward to custom error page
                try:
                    self.forward(trans, url)
                except EndResponse:
                    pass
                except ConnectionAbortedError as err:
                    trans.setError(err)
                except Exception:
                    # If the custom error page itself throws an exception,
                    # display the new exception instead of the original one,
                    # so we notice that something is broken here.
                    url = None
                if url:
                    return  # error has already been handled
            if isHTTPException:
                # display standard http error page
                trans.response().displayError(err)
            else:
                # standard error handling
                if (self.setting('EnterDebuggerOnException')
                        and sys.stdin.isatty()):
                    import pdb
                    pdb.post_mortem(sys.exc_info()[2])
                self.handleExceptionInTransaction(sys.exc_info(), trans)

    @staticmethod
    def runTransactionViaServlet(servlet, trans):
        """Execute the transaction using the servlet.

        This is the `awake`/`respond`/`sleep` sequence of calls, or if
        the servlet supports it, a single `runTransaction` call (which is
        presumed to make the awake/respond/sleep calls on its own). Using
        `runTransaction` allows the servlet to override the basic call
        sequence, or catch errors from that sequence.

        Called by `runTransaction`.
        """
        trans.setServlet(servlet)
        if hasattr(servlet, 'runTransaction'):
            servlet.runTransaction(trans)
        else:
            # For backward compatibility (Servlet.runTransaction implements
            # this same sequence of calls, but by keeping it in the servlet
            # it's easier for the servlet to catch exceptions).
            try:
                trans.awake()
                trans.respond()
            finally:
                trans.sleep()

    def forward(self, trans, url):
        """Forward the request to a different (internal) URL.

        The transaction's URL is changed to point to the new servlet,
        and the transaction is simply run again.

        Output is _not_ accumulated, so if the original servlet had any output,
        the new output will _replace_ the old output.

        You can change the request in place to control the servlet you are
        forwarding to -- using methods like `HTTPRequest.setField`.
        """
        # Reset the response to a "blank slate"
        trans.response().reset()
        # Include the other servlet
        self.includeURL(trans, url)
        # Raise an exception to end processing of this request
        raise EndResponse

    def callMethodOfServlet(self, trans, url, method, *args, **kw):
        """Call method of another servlet.

        Call a method of the servlet referred to by the URL. Calls sleep()
        and awake() before and after the method call. Or, if the servlet
        defines it, then `runMethodForTransaction` is used (analogous to the
        use of `runTransaction` in `forward`).

        The entire process is similar to `forward`, except that instead of
        `respond`, `method` is called (`method` should be a string,
        `*args` and `**kw` are passed as arguments to that method).
        """
        # store current request and set the new URL
        request = trans.request()
        request.push(trans.servlet(),
                     self.resolveInternalRelativePath(trans, url))
        # get new servlet
        servlet = self.rootURLParser().findServletForTransaction(trans)
        trans.setServlet(servlet)
        # call method of included servlet
        if hasattr(servlet, 'runMethodForTransaction'):
            result = servlet.runMethodForTransaction(
                trans, method, *args, **kw)
        else:
            servlet.awake(trans)
            result = getattr(servlet, method)(*args, **kw)
            servlet.sleep(trans)
        # return new servlet to its pool
        self.returnServlet(servlet)
        # restore current request
        trans.setServlet(request.pop())
        return result

    def includeURL(self, trans, url):
        """Include another servlet.

        Include the servlet given by the URL. Like `forward`,
        except control is ultimately returned to the servlet.
        """
        # store current request and set the new URL
        request = trans.request()
        request.push(trans.servlet(),
                     self.resolveInternalRelativePath(trans, url))
        # get new servlet
        servlet = self.rootURLParser().findServletForTransaction(trans)
        trans.setServlet(servlet)
        # run the included servlet
        try:
            servlet.runTransaction(trans)
        except EndResponse:
            # we interpret an EndResponse in an included page to mean
            # that the current page may continue processing
            pass
        # return new servlet to its pool
        self.returnServlet(servlet)
        # restore current request
        trans.setServlet(request.pop())

    @staticmethod
    def resolveInternalRelativePath(trans, url):
        """Return the absolute internal path.

        Given a URL, return the absolute internal URL.
        URLs are assumed relative to the current URL.
        Absolute paths are returned unchanged.
        """
        if not url.startswith('/'):
            origDir = trans.request().urlPath()
            if not origDir.endswith('/'):
                origDir = os.path.dirname(origDir)
                if not origDir.endswith('/'):
                    origDir += '/'
            url = origDir + url
        # deal with . and .. in the path
        parts = []
        for part in url.split('/'):
            if parts and part == '..':
                parts.pop()
            elif part != '.':
                parts.append(part)
        return '/'.join(parts)

    @staticmethod
    def returnServlet(servlet):
        """Return the servlet to its pool."""
        servlet.close()

    def errorPage(self, errorClass):
        """Get the error page url corresponding to an error class."""
        if errorClass.__name__ in self._errorPage:
            return self._errorPage[errorClass.__name__]
        if errorClass is not Exception:
            for errorClass in errorClass.__bases__:
                url = self.errorPage(errorClass)
                if url:
                    return url

    def handleException(self):
        """Handle exceptions.

        This should only be used in cases where there is no transaction object,
        for example if an exception occurs when attempting to save a session
        to disk.
        """
        self._exceptionHandlerClass(self, None, sys.exc_info())

    def handleExceptionInTransaction(self, excInfo, trans):
        """Handle exception with info.

        Handles exception `excInfo` (as returned by `sys.exc_info()`)
        that was generated by `transaction`. It may display the exception
        report, email the report, etc., handled by
        `ExceptionHandler.ExceptionHandler`.
        """
        request = trans.request()
        editlink = f'{request.scriptName()}/Admin/EditFile' if self.setting(
            'IncludeEditLink') else None
        self._exceptionHandlerClass(
            self, trans, excInfo, dict(editlink=editlink))

    def rootURLParser(self):
        """Accessor: the Root URL parser.

        URL parsing (as defined by subclasses of `URLParser.URLParser`)
        starts here. Other parsers are called in turn by this parser.
        """
        return self._rootURLParser

    def hasContext(self, name):
        """Checks whether context `name` exist."""
        return name in self._rootURLParser._contexts

    def addContext(self, name, path):
        """Add a context by named `name`, rooted at `path`.

        This gets imported as a package, and the last directory
        of `path` does not have to match the context name.
        (The package will be named `name`, regardless of `path`).

        Delegated to `URLParser.ContextParser`.
        """
        self._rootURLParser.addContext(name, path)

    @staticmethod
    def addServletFactory(factory):
        """Add a ServletFactory.

        Delegated to the `URLParser.ServletFactoryManager` singleton.
        """
        URLParser.ServletFactoryManager.addServletFactory(factory)

    def contexts(self):
        """Return a dictionary of context-name: context-path."""
        return self._rootURLParser._contexts

    _exceptionReportAttrNames = [
        'webwareVersion', 'webwarePath', 'serverSidePath', 'contexts']

    def writeExceptionReport(self, handler):
        """Write extra information to the exception report.

        See `ExceptionHandler` for more information.
        """
        handler.writeTitle(self.__class__.__name__)
        handler.writeAttrs(self, self._exceptionReportAttrNames)

    @staticmethod
    def removePathSession(trans):
        """Remove a possible session identifier from the path."""
        request = trans.request()
        # Try to get automatic path session:
        # If UseAutomaticPathSessions is enabled in Application.config,
        # Application redirects the browser to a URL with SID in path:
        # http://gandalf/a/_SID_=2001080221301877755/Examples/
        # The _SID_ part is extracted and removed from path here.
        # Note that We do not check for the exact name of the field
        # here because it may be context dependent.
        p = request._pathInfo
        if p:
            p = p.split('/', 2)
            if len(p) > 1 and not p[0] and '=' in p[1]:
                s = p[1]
                request._pathSID = s.split('=', 1)
                s += '/'
                del p[1]
                env = request.environ()
                request._pathInfo = env['PATH_INFO'] = '/'.join(p)
                for v in ('REQUEST_URI', 'PATH_TRANSLATED'):
                    if v in env:
                        env[v] = env[v].replace(s, '', 1)
            else:
                request._pathSID = None
        else:
            request._pathSID = None

    def handlePathSession(self, trans):
        """Handle the session identifier that has been found in the path."""
        request = trans.request()
        if request._pathSID:
            sessionName = self.sessionName(trans)
            if request._pathSID[0] == sessionName:
                if request.hasCookie(sessionName):
                    self.handleUnnecessaryPathSession(trans)
                request.cookies()[sessionName] = request._pathSID[1]
            else:
                if self._autoPathSessions:
                    if not request.hasCookie(sessionName):
                        self.handleMissingPathSession(trans)
        else:
            if self._autoPathSessions:
                if not request.hasCookie(self.sessionName(trans)):
                    self.handleMissingPathSession(trans)

    def handleMissingPathSession(self, trans):
        """Redirect requests without session info in the path.

        If UseAutomaticPathSessions is enabled in Application.config
        we redirect the browser to an absolute url with SID in path
        http://gandalf/a/_SID_=2001080221301877755/Examples/
        _SID_ is extracted and removed from path in HTTPRequest.py

        This is for convenient building of webapps that must not
        depend on cookie support.

        Note that we create an absolute URL with scheme and hostname
        because otherwise IIS will only cause an internal redirect.
        """
        request = trans.request()
        url = '{}://{}{}/{}={}{}{}'.format(
            request.scheme(), request.hostAndPort(), request.servletPath(),
            self.sessionName(trans), trans.session().identifier(),
            request.pathInfo(), request.extraURLPath() or '')
        if request.queryString():
            url += '?' + request.queryString()
        if self.setting('Debug')['Sessions']:
            print('>> [sessions] handling UseAutomaticPathSessions,'
                  ' redirecting to', url)
        trans.response().sendRedirect(url)
        raise EndResponse

    def handleUnnecessaryPathSession(self, trans):
        """Redirect request with unnecessary session info in the path.

        This is called if it has been determined that the request has a path
        session, but also cookies. In that case we redirect to eliminate the
        unnecessary path session.
        """
        request = trans.request()
        url = '{}://{}{}{}{}'.format(
            request.scheme(), request.hostAndPort(), request.servletPath(),
            request.pathInfo(), request.extraURLPath() or '')
        if request.queryString():
            url += '?' + request.queryString()
        if self.setting('Debug')['Sessions']:
            print('>> [sessions] handling unnecessary path session,'
                  ' redirecting to', url)
        trans.response().sendRedirect(url)
        raise EndResponse

    # endregion Request Dispatching

    # region Plug-in loading

    def plugIns(self):
        """Return a dictionary of the plug-ins loaded by the application.

        Each plug-in is a PlugIn object with an underlying Python package.
        """
        return self._plugIns

    def plugIn(self, name, default=NoDefault):
        """Return the plug-in with the given name."""
        if default is NoDefault:
            return self._plugIns[name]
        return self._plugIns.get(name, default)

    def loadPlugIn(self, name, module):
        """Load and return the given plug-in.

        May return None if loading was unsuccessful (in which case this method
        prints a message saying so). Used by `loadPlugIns` (note the **s**).
        """
        return self._plugInLoader(name, module)

    def loadPlugIns(self):
        """Load all plug-ins.

        A plug-in allows you to extend the functionality of Webware without
        necessarily having to modify its source. Plug-ins are loaded by
        Application at startup time, just before listening for requests.
        See the docs in `PlugIn` for more info.
        """
        self._plugInLoader = loader = PlugInLoader(self)
        self._plugIns = loader.loadPlugIns(self.setting('PlugIns'))

    # endregion Plug-in loading

    # region WSGI interface

    def __call__(self, environ, start_response):
        """The WSGI application callable"""
        verbose = self._verbose
        requestDict = dict(environ=environ)

        requestID = self._requestID
        self._requestID = requestID + 1
        startTime = time()
        requestTime = startTime
        requestDict['requestID'] = requestID
        requestDict['time'] = requestTime
        requestDict['format'] = 'CGI'
        if verbose:
            uri = requestURI(environ)
            if not self._silentURIs or not self._silentURIs.search(uri):
                requestDict['verbose'] = True
                requestTime = localtime(requestTime)[:6]
                fmt = '{:5d}  {:4d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}  {}'
                print(fmt.format(requestID, *requestTime, uri))

        requestDict['input'] = environ['wsgi.input']

        streamOut = WSGIStreamOut(
            start_response, bufferSize=self._responseBufferSize,
            useWrite=self._wsgiWrite, encoding=self._outputEncoding)
        transaction = self.dispatchRawRequest(requestDict, streamOut)
        try:
            streamOut.close()
            aborted = False
        except ConnectionAbortedError:
            aborted = True

        if verbose and requestDict.get('verbose'):
            endTime = time()
            duration = round((endTime - startTime) * 1000)
            if aborted:
                error = '* connection aborted *'
            else:
                error = requestURI(environ)
            fmt = '{:5d}  {:16.0f} ms  {}\n'
            print(fmt.format(requestID, duration, error))

        transaction._application = None
        transaction.die()
        del transaction

        return streamOut.iterable()

    # endregion WSGI interface



"""The RoleUser class."""

from .User import User


class RoleUser(User):
    """In conjunction with Role, provides role-based users and security.

    See the doc for playsRole() for an example.

    Note that this class plays nicely with both Role and HierRole,
    e.g., no "HierRoleUser" is needed when making use of HierRoles.

    See also:
      * class Role
      * class HierRole
    """

    # region Init

    def __init__(self, manager=None, name=None, password=None):
        User.__init__(self, manager, name, password)
        self._roles = []
        self._rolesByName = {}

    # endregion Init

    # region Accessing roles

    def roles(self):
        """Return a direct list of the user's roles.

        Do not modify.
        """
        return self._roles

    def setRoles(self, listOfRoles):
        """Set all the roles for the user.

        Each role in the list may be a valid role name or a Role object.

        Implementation note: depends on addRoles().
        """
        self._roles = []
        self.addRoles(listOfRoles)

    def addRoles(self, listOfRoles):
        """Add additional roles for the user.

        Each role in the list may be a valid role name or a Role object.
        """
        start = len(self._roles)
        self._roles.extend(listOfRoles)

        # Convert names to role objects and update self._rolesByName
        index = start
        numRoles = len(self._roles)
        while index < numRoles:
            role = self._roles[index]
            if isinstance(role, str):
                role = self._manager.roleForName(role)
                self._roles[index] = role
            self._rolesByName[role.name()] = role
            index += 1

    def playsRole(self, roleOrName):
        """Check whether the user plays the given role.

        More specifically, if any of the user's roles return true for
        role.playsRole(otherRole), this method returns True.

        The application of this popular method often looks like this:
            if user.playsRole('admin'):
                self.displayAdminMenuItems()
        """
        if isinstance(roleOrName, str):
            roleOrName = self._manager.roleForName(roleOrName)
        return any(role.playsRole(roleOrName) for role in self._roles)

    # endregion Accessing roles



from time import sleep

from Page import Page


class PushServlet(Page):
    """Pushing Content Demo.

    This is a test servlet for the buffered output streams of the application.

    Server push is a simple way of sending file updates to a browser.
    It is not true streaming, though it can feel a lot like it when done well.

    The browser needs to support "multipart/x-mixed-replace" for this to work.
    Some browsers like MSIE do not support this, and others like Chrome have
    restricted it to images or dropped support for this feature.

    The WSGI server must also support this.

    This demo has been confirmed to work with Firefox 68 and waitress 1.3.
    """

    _boundary = ("if-you-see-this-your-browser-then"
                 "-it-does-not-support-multipart/x-mixed-replace")

    def respond(self, transaction):
        response = self.response()
        # this isn't necessary, but it's here as an example:
        response.streamOut().setAutoCommit()
        # send the initial header:
        self.initialHeader()
        self.sendBoundary()
        response.flush()
        # push new content 4 times:
        for i in range(4):
            self.pushHTML(i)
            self.sendBoundary()
            # send the currently buffered output now:
            response.flush()
            sleep(5 if i else 15)

    def initialHeader(self):
        setHeader = self.response().setHeader
        setHeader("Content-Type",
                  "multipart/x-mixed-replace; boundary=" + self._boundary)
        setHeader("X-Accel-Buffering", "no")

    def sendBoundary(self):
        self.write('--' + self._boundary + '\r\n')

    def pushHTML(self, count):
        self.write("Content-Type: text/html\r\n\r\n")
        wr = self.writeln
        wr('<html><body style="margin:8pt;"><div style='
           '"background-color:#eef;padding:8pt 16pt;border:2px solid blue">')
        wr('<h1>Pushing Content Demo</h1>')
        if count:
            wr('<h3>This page has been replaced'
               f' <strong style="color:#339">{count:d}</strong>'
               f" time{'' if count ==1 else 's'}.</h3>")
            if count == 3:
                wr('<p>Stopped pushing contents.</p>')
            else:
                wr('<p>Next content will be pushed'
                   ' <strong>in 5</strong> seconds.</p>')
        else:
            wr('<p>This servlet will try to replace the content'
               ' <strong>in 15 seconds</strong>.</p>')
        if not count or count == 3:
            wr('<h4>Note:</h4>')
            if count == 3:
                wr("<p>If you didn't get output for the last 30 seconds, "
                   'pushing contents is not supported.</p>')
            wr('<p>The browser must support the <code>x-mixed-replace</code>'
               ' content type. You should try with Firefox first.</p>')
        wr('</div></body></html>')



"""ParamFactory.py

A factory for creating cached, parametrized class instances.
"""

from threading import Lock


class ParamFactory:

    def __init__(self, klass, **extraMethods):
        self.lock = Lock()
        self.cache = {}
        self.klass = klass
        for name, func in list(extraMethods.items()):
            setattr(self, name, func)

    def __call__(self, *args):
        self.lock.acquire()
        if args in self.cache:
            self.lock.release()
            return self.cache[args]
        value = self.klass(*args)
        self.cache[args] = value
        self.lock.release()
        return value

    def allInstances(self):
        return list(self.cache.values())
